/*
TODO:
 


*/
#define PCB_1_0
//#define PCB_alpha

#include <avr/io.h>
#include "tinybass.h"
#include "timer.h"

/***********************
 * Define some setting *
 ***********************/

#define STARTNOTE 9

#define STARTPWM 2
#define PWM_UPDATE_TIME 20
#define PWM_MIN (Note[thisnote]/STARTPWM)
#define PWM_MAX Note[thisnote]

#define F_MIN (TMROFFSET+1)
#define F_MAX 255
#define F_MOD_TIME 20
#define FILTERSTEPS (256-TMROFFSET)

#define MAX_TIME 20000

#define MAXNOTES 32

#define TIME_OF_BUTTON_HOLDS 2000

/********************
 * GLOBAL VARIABLES *
 ********************/

long time=300;

short melody[MAXNOTES];
short melody_position=0;
short melody_length=MAXNOTES;
short tune[MAXNOTES];

/*Parameters set by the rotary*/
int p_pwm=2;
int p_filter=5;
short p_tune=0;
unsigned char p_melody=1;
int p_extclock=8;
short tap_tempo_mode=1;


/*******************
 * PIN DEFINITIONS *
 *******************/

#ifdef PCB_alpha
#define MOD_in PA4
#define TAP_in PA2
#define TUNE_in PA3
#endif

#ifdef PCB_1_0
#define MOD_in PA2
#define TAP_in PA4
#define TUNE_in PA3
#endif

#define PWM_output PA5
#define PWM_output_init DDRA |= (1<<PWM_output)

#define CLK_input PA7
#define CLK_pin_init DDRA &= ~(1<<CLK_input)
#define CLK_IR_RECV (!(PINA & (1<<CLK_input)))

#define TAP_pin_init (DDRA &= ~(1<<TAP_in)) && (PORTA |= (1<<TAP_in))
#define TAP_TEMPO_BUTTON (!(PINA & (1<<TAP_in)))

#define TUNE_pin_init (DDRA &= ~(1<<TUNE_in)) && (PORTA |= (1<<TUNE_in))
#define TUNE_BUTTON (!(PINA & (1<<TUNE_in)))

#define MOD_pin_init (DDRA &= ~(1<<MOD_in)) && (PORTA |= (1<<MOD_in))
#define MOD_BUTTON (!(PINA & (1<<MOD_in)))


#define ROTARY_CW_pin PB0
#define ROTARY_CCW_pin PB1
#define ROTARY_init DDRB &= ~((1<<ROTARY_CW_pin) | (1<<ROTARY_CCW_pin))
#define ROTARY_pullup_init PORTB |= (1<<ROTARY_CW_pin) | (1<<ROTARY_CCW_pin)
#define ROTARY_CW (!(PINB & (1<<ROTARY_CW_pin)))
#define ROTARY_CCW (!(PINB & (1<<ROTARY_CCW_pin)))

#define IR1_pin PA0
#define IR1_pin_init DDRA |= (1<<IR1_pin)
#define IR1_on PORTA |=(1<<IR1_pin)
#define IR1_off PORTA &= ~(1<<IR1_pin)

#define IR2_pin PA1
#define IR2_pin_init DDRA |= (1<<IR2_pin)
#define IR2_on PORTA |=(1<<IR2_pin)
#define IR2_off PORTA &= ~(1<<IR2_pin)

#define LED_pin PB2
#define LED_pin_init DDRB |= (1<<LED_pin)
#define LED_on PORTA |=(1<<LED_pin)
#define LED_off PORTA &= ~(1<<LED_pin)

void init_pins(void){
	IR1_pin_init;
	IR2_pin_init;
	LED_pin_init;
	CLK_pin_init;
	TAP_pin_init;
	TUNE_pin_init;
	MOD_pin_init;
	ROTARY_init;
	ROTARY_pullup_init;
}


/*******
 * PWM *
 *******/

void init_pwm(void){

	//set PWM pin to output
	PWM_output_init;

	//set WGM (Waveform Generation Mode) to 15 = fast PWM mode with OCR1A as TOP
	//set COM1B to 3  = set OC1B pin on compare match, clear OC1B pin at BOTTOM
	//set CS (clock pre-scaling) to 2 = prescale FOSC/8 (equals 1MHz)

	TCCR1A = (1<<WGM10) | (1<<WGM11) | (1<<COM1B0) | (1<<COM1B1);
	TCCR1B = (1<<WGM13) | (1<<WGM12) | (0<<CS10) | (1<<CS11) | (0<<CS12);

	OCR1A = STARTNOTE;
	OCR1B = STARTNOTE/STARTPWM;
}



/*************
 *   TEMPO   *
 *************/

inline short set_tempo(long t){
	if (t>30) {time=t;return(t);}
	else return(0);
}

short tap_tempo(void){
	static unsigned char ext_clock_divide=0;
	static long note_timer=0;
	static long time_of_tap=0;
	short play_the_next_note=0;
	static long ir_pulse=0;
	long now=gettmrms();
	static short  IR_down=0;
	static long button_down=0;

	if (tap_tempo_mode){

		if (now>=(time+note_timer)){
		 	IR1_on;
	//		IR2_on;

				if (++ext_clock_divide>=(p_extclock>>3)){
					play_the_next_note=1;
					ext_clock_divide=0;
					IR2_on;
				}

			ir_pulse=now;

	//		play_the_next_note=1;
			note_timer=now;
		}

		//After 40ms, turn off the IR SEND LEDs
		if (now>=(ir_pulse+40)) {
			IR1_off;
			IR2_off;
		}
		

		/*special flag to indicate TAP_BUTTON is being held down while
		the user modifies a parameter: so tap_tempo() function should not interpret it as a
		switch to ext clock mode*/
		if (tap_tempo_mode==5) {
			tap_tempo_mode=4;
			time_of_tap=now;
		}



		if (!MOD_BUTTON && !TUNE_BUTTON){
			switch (tap_tempo_mode){
				case (4): //mode 4 is button down on second tap
				case (2): //mode 2 is button down on first tap
					if (!TAP_TEMPO_BUTTON){ //wait for the button to be released
						tap_tempo_mode++; 
						if (tap_tempo_mode>4) tap_tempo_mode=1;
					} else {
						if (now>=(time_of_tap+TIME_OF_BUTTON_HOLDS)){ 
											//if you hold the button down for more than 2 second...
							button_down=time_of_tap;			//then we go to external clock
							tap_tempo_mode=0;
						}
					}
					break;


				case (3): //mode 3 is button up after first tap
					if (TAP_TEMPO_BUTTON){ //...waiting to goto mode 4 when button pressed
						set_tempo(now-time_of_tap);

						tap_tempo_mode=4;
						time_of_tap=gettmrms();

					} else {

						//If too much time passes between taps,
						//then reset the tap timer to mode 1 (waiting for first tap)
						if(now >= (time_of_tap+MAX_TIME)){
							time_of_tap=0;
							tap_tempo_mode=1;
						}
					}
					break;

				case (1): //mode 1 is just playing (waiting for first tap)
					if (TAP_TEMPO_BUTTON){ //...waiting to goto mode 2 when button pressed

						if ((now - time_of_tap) < (time*2))	{
							if (set_tempo(now-time_of_tap)) play_the_next_note=1;
						}

						time_of_tap=gettmrms();
						tap_tempo_mode=2; 
					}
					break;
			}
		}

	} else {
	// EXT CLOCK

		if (CLK_IR_RECV) {
			if (!IR_down){
				IR1_on;
				IR_down=1;

				if (++ext_clock_divide>=(p_extclock>>3)){
					play_the_next_note=1;
					ext_clock_divide=0;
					if (set_tempo(now-time_of_tap)){	
						time_of_tap=now;
					}
					IR2_on;
				}
			}
		}
		else {
			if (IR_down){ //falling edge
				IR_down=0;
				IR1_off;
				IR2_off;
			}

		}


		if (TAP_TEMPO_BUTTON){
			if (!button_down) button_down=now;
		}

		if (!TAP_TEMPO_BUTTON){
			if ((now-button_down)<TIME_OF_BUTTON_HOLDS)
				tap_tempo_mode=1;					//go back to tap_tempo_mode
			else
				button_down=0; //when the button is up, reset the flag

		}

	}

	return(play_the_next_note);


}


/*********************
 * Generate_Melody() *
 *********************/

#define MELS_IN_A_SHELL 16


void Generate_Melody(int seed){

	short b0,b1,bit,i,write_pos;

	if (p_melody==-1) melody_length=0;
	else melody_length = ((p_melody / MELS_IN_A_SHELL) + 1) * 8;

//	melody_length = melody_length % MAXNOTES;

	b0=((p_melody>>0) & (MELS_IN_A_SHELL-1)) + 1; 


	for(i=0;i<MAXNOTES;i++) melody[i]=STARTNOTE;
	
	write_pos = 0;
	for (bit=0;bit<b0;bit++){

		b1=bit & 0b111;

		for (i=0;i<(melody_length / ((b1)+2));i++){

			melody[write_pos]+=(b1+1);
			if (melody[write_pos]> HIGHEST_NOTE) melody[write_pos]-=HIGHEST_NOTE;
			write_pos+=(b1+2);
			write_pos= (write_pos % melody_length) & (MAXNOTES-1);

		}
	}

	for(i=0;i<MAXNOTES;i++) if (melody[i]==STARTNOTE) melody[i]=0;

	if (melody_position>melody_length) melody_position=melody_position % melody_length;

}

void init_notes(void){
	short i;

	p_melody=1;
	Generate_Melody(1);

	p_tune=0;
	p_pwm=2;
	p_filter=5;
	p_tune=0;
	p_extclock=8;

	for(i=0;i<MAXNOTES;i++){
		tune[i]=p_tune;
	}

	tap_tempo_mode=1;
	time=300;
	melody_position=0;
}

/****************
 * TuneMelody() *
 ****************/
void TuneMelody(unsigned char n){
	short i;
	if (n>HIGHEST_NOTE) n=HIGHEST_NOTE;
	for (i=0;i<MAXNOTES;i++)
		tune[i]=n;
}


short changed(int a, int b, short amount){
	int c=a-b;
	if (c<0) c=-c;
	if (c>amount) return(c);
	else return(0);
}



/*************
 * Main()    *
 *************/

int main(void){
	short play_the_next_note=0;
	short thisnote=0;

	long pwm_timer=0;
	unsigned int pwm_step=0;
	short stop_pwm=0;
	unsigned int t_OCR1B=0;
	int t_OCR0A=0;

	long filter_timer=0;
	long filter_step=0;

	short greycode=0;
	short rotary_changed=0;

	short mel12=0;

	long manual_reset_down=0;

	/** Initialize **/

	init_pwm();
	inittimer();
	init_notes();
	init_pins();
	TuneMelody(p_tune);


	/** Main loop **/
	while(1){ 

		/** read the rotary **/
				
	
		if ((greycode==0b11) && !ROTARY_CW){ //greycode was 0b11, now its 0b01 so rotary went down
			rotary_changed+=1;
			greycode=0b01;
		}

		else if ((greycode==0b11) && !ROTARY_CCW){ //greycode was 0b11, now its 0b10 so rotary went up
			rotary_changed-=1;
			greycode=0b10;
		}

		else if (ROTARY_CW && ROTARY_CCW) //greycode is 11
			greycode=0b11;

		else greycode=0b00;

		if (rotary_changed!=0){
			if (TAP_TEMPO_BUTTON){
				if (tap_tempo_mode) tap_tempo_mode=5;
				p_extclock-=rotary_changed;
				if (p_extclock<8) p_extclock=8;
				if (p_extclock>256) p_extclock=256;

			} else if (TUNE_BUTTON && MOD_BUTTON){
				p_pwm-=rotary_changed;


			} else if (TUNE_BUTTON){
				p_tune+=rotary_changed;
				if (p_tune<=(-1*STARTNOTE))
					p_tune=(-1*STARTNOTE)+1;
				if (p_tune>=(HIGHEST_NOTE-STARTNOTE))
					p_tune=HIGHEST_NOTE-STARTNOTE-1;

			//	TuneMelody(p_tune);

				thisnote = p_tune+ STARTNOTE+1;
				OCR1A = Note[thisnote];
				t_OCR1B= PWM_MIN;
				OCR1B = t_OCR1B;
				pwm_timer=gettmrms();

			
			} else if (MOD_BUTTON){
				p_filter+=rotary_changed;
				
			} else {
				p_melody+=rotary_changed;
				Generate_Melody(p_melody);
			}

			rotary_changed=0;
		} else {
			if (MOD_BUTTON && TAP_TEMPO_BUTTON && TUNE_BUTTON){
				if (!manual_reset_down)	{
					manual_reset_down=gettmrms();
				} else if ((gettmrms()-manual_reset_down)>TIME_OF_BUTTON_HOLDS){
					init_notes();

				}
			} else manual_reset_down=0;

		}



		/** Read the tempo **/

		play_the_next_note=tap_tempo();

		long now=gettmrms();

		if (play_the_next_note){
			play_the_next_note=0;
			


			if (TUNE_BUTTON && !MOD_BUTTON)
				thisnote=STARTNOTE+1;
			else 
				thisnote = melody[melody_position];

			//Tune the note up, except for note 0, which is "rest"
			if (thisnote>0) thisnote+=p_tune; 

			//Range check: make sure we stay within usable note range.
			if (thisnote>=HIGHEST_NOTE) thisnote=HIGHEST_NOTE-1;

			//Set the pitch to the current note (OCR1A is pitch)
			OCR1A = Note[thisnote];


			//Initialize the note's PWM (OCR1B is PWM)

			if (p_pwm>0) t_OCR1B = PWM_MIN;
			else t_OCR1B = PWM_MAX;
			OCR1B = t_OCR1B;

			long steps=time/p_pwm;
			pwm_step=(PWM_MAX-PWM_MIN)/steps;
			pwm_timer=now;
			stop_pwm=0;
			
			long fsteps=time/p_filter;
			filter_step=(F_MAX-F_MIN)/fsteps;
			if (p_filter>0) t_OCR0A=F_MIN;
			else if (p_filter<0) t_OCR0A=F_MAX;
			else t_OCR0A=(F_MAX-F_MIN)/2 + F_MIN;
			OCR0A=t_OCR0A;
			filter_timer=now;

			//increment the melody position, and loop it around the the beginning when we hit the end
			
			if (++melody_position>=melody_length){
				melody_position=0;
				if (++mel12==2) {p_melody++;mel12=0;Generate_Melody(p_melody);}
			}
		}
		else
		{

			if (!stop_pwm && (now>=(pwm_timer+PWM_UPDATE_TIME))){
				pwm_timer+=PWM_UPDATE_TIME;
				t_OCR1B += pwm_step;

				if (t_OCR1B>PWM_MAX) {stop_pwm=1;t_OCR1B=PWM_MAX;}
				else if (t_OCR1B<PWM_MIN) t_OCR1B+=PWM_MAX-PWM_MIN;

				OCR1B=t_OCR1B;

			}
			
			if (now>=(filter_timer+F_MOD_TIME)){
				filter_timer+=F_MOD_TIME;
				t_OCR0A += filter_step;

				if (t_OCR0A > F_MAX) t_OCR0A -= (F_MAX-F_MIN);
				if (t_OCR0A < F_MIN) t_OCR0A += (F_MAX-F_MIN);

				OCR0A = t_OCR0A;

			}

			
		}


	}

}





